!--------------------------------------------------------------------------!
! The Phantom Smoothed Particle Hydrodynamics code, by Daniel Price et al. !
! Copyright (c) 2007-2019 The Authors (see AUTHORS)                        !
! See LICENCE file for usage and distribution conditions                   !
! http://phantomsph.bitbucket.io/                                          !
!--------------------------------------------------------------------------!
!+
!  MODULE: testkdtree
!
!  DESCRIPTION:
!   This module performs unit tests of the kdtree module
!   The tests here are specific to the tree, some general
!   tests of neighbour finding are done in test_link
!
!  REFERENCES: None
!
!  OWNER: Daniel Price
!
!  $Id: 48c8905c33e489dfd26b003387e5b96d3afd5a32 $
!
!  RUNTIME PARAMETERS: None
!
!  DEPENDENCIES: dim, io, kdtree, kernel, linklist, part, testutils,
!    timing, unifdis
!+
!--------------------------------------------------------------------------
module testkdtree
 implicit none
 public :: test_kdtree

 private

contains

subroutine test_kdtree(ntests,npass)
 use dim,       only:maxp
 use io,        only:id,master,iverbose
 use linklist,  only:ifirstincell,ncells,node
 use part,      only:npart,xyzh,hfact,massoftype,igas,maxphase,iphase,isetphase
 use kernel,    only:hfact_default
 use kdtree,    only:maketree,revtree,kdnode,empty_tree
 use unifdis,   only:set_unifdis
 use testutils, only:checkvalbuf,checkvalbuf_end,update_test_scores
 use timing,    only:print_time,getused
 integer, intent(inout) :: ntests,npass
 logical :: test_revtree, test_all
 integer :: i,nfailed(12),nchecked(12)
 real    :: psep,tol,errmax(12)
 real(4) :: t2,t1
 type(kdnode), allocatable :: old_tree(:)

 test_all = .false.
 test_revtree = .false.
 iverbose = 2

 if (id==master) write(*,"(a,/)") '--> TESTING KDTREE'

 if (test_revtree .or. test_all) then
    if (id==master) write(*,"(/,a)") '--> testing revtree routine'
    !
    ! set up a random particle distribution
    !
    psep = 1./64.
    hfact = hfact_default
    npart = 0
    call set_unifdis('random',id,master,-0.5,0.5,-0.5,0.5,-0.5,0.5,psep,hfact,npart,xyzh)
    massoftype(igas) = 1000./npart
    if (maxphase==maxp) iphase(:) = isetphase(igas,iactive=.true.)

    !
    ! call maketree to build the tree
    !
    call empty_tree(node)
    call cpu_time(t1)
    call maketree(node,xyzh,npart,3,ifirstincell,ncells)
    call cpu_time(t2)
    call print_time(t2-t1,'maketree completed in')
    !
    ! now save the tree structure
    !
    allocate(old_tree(ncells))
    old_tree(1:ncells) = node(1:ncells)

    !
    ! erase all information in the existing tree except the structure
    !
    do i=1,int(ncells)
#ifdef GRAVITY
       node(i)%xcen(:) = 0.
#endif
       node(i)%size    = 0.
       node(i)%hmax    = 0.
#ifdef GRAVITY
       node(i)%mass    = 0.
       node(i)%quads(:)= 0.
#endif
    enddo

    !
    ! call revtree to rebuild
    !
    call cpu_time(t1)
    call revtree(node,xyzh,ifirstincell,ncells)
    call cpu_time(t2)
    call print_time(t2-t1,'revtree completed in')

    !
    ! check that the revised tree matches the tree built
    !
    nfailed(:)  = 0
    nchecked(:) = 0
    errmax(:)   = 0.
    tol = 1.8e-13 !epsilon(0.)
    do i=1,int(ncells)
       ! if (ifirstincell(i) /= 0) then
       call checkvalbuf(node(i)%xcen(1),old_tree(i)%xcen(1),tol,'x0',nfailed(1),nchecked(1),errmax(1))
       call checkvalbuf(node(i)%xcen(2),old_tree(i)%xcen(2),tol,'y0',nfailed(2),nchecked(2),errmax(2))
       call checkvalbuf(node(i)%xcen(3),old_tree(i)%xcen(3),tol,'z0',nfailed(3),nchecked(3),errmax(3))
!       call checkvalbuf(node(i)%size,old_tree(i)%size,tol,'size',nfailed(4),nchecked(4),errmax(4))
       call checkvalbuf((node(i)%size + tol >= old_tree(i)%size),.true.,'size',nfailed(4),nchecked(4))
       call checkvalbuf(node(i)%hmax,old_tree(i)%hmax,tol,'hmax',nfailed(5),nchecked(5),errmax(5))
#ifdef GRAVITY
       call checkvalbuf(node(i)%mass,old_tree(i)%mass,tol,'mass',nfailed(6),nchecked(6),errmax(6))
       call checkvalbuf(node(i)%quads(1),old_tree(i)%quads(1),tol,'qxx',nfailed(7),nchecked(7),errmax(7))
       call checkvalbuf(node(i)%quads(2),old_tree(i)%quads(2),tol,'qxy',nfailed(8),nchecked(8),errmax(8))
       call checkvalbuf(node(i)%quads(3),old_tree(i)%quads(3),tol,'qxz',nfailed(9),nchecked(9),errmax(9))
       call checkvalbuf(node(i)%quads(4),old_tree(i)%quads(4),tol,'qyy',nfailed(10),nchecked(10),errmax(10))
       call checkvalbuf(node(i)%quads(5),old_tree(i)%quads(5),tol,'qyz',nfailed(11),nchecked(11),errmax(11))
       call checkvalbuf(node(i)%quads(6),old_tree(i)%quads(6),tol,'qzz',nfailed(12),nchecked(12),errmax(12))
#endif
       ! endif
    enddo
    call checkvalbuf_end('x0',nchecked(1),nfailed(1),errmax(1),tol)
    call checkvalbuf_end('y0',nchecked(2),nfailed(2),errmax(2),tol)
    call checkvalbuf_end('z0',nchecked(3),nfailed(3),errmax(3),tol)
    call checkvalbuf_end('size',nchecked(4),nfailed(4),errmax(4),tol)
    call checkvalbuf_end('hmax',nchecked(5),nfailed(5),errmax(5),tol)
#ifdef GRAVITY
    call checkvalbuf_end('mass',nchecked(6),nfailed(6),errmax(6),tol)
    call checkvalbuf_end('qxx',nchecked(7),nfailed(7),errmax(7),tol)
    call checkvalbuf_end('qxy',nchecked(8),nfailed(8),errmax(8),tol)
    call checkvalbuf_end('qxz',nchecked(9),nfailed(9),errmax(9),tol)
    call checkvalbuf_end('qyy',nchecked(10),nfailed(10),errmax(10),tol)
    call checkvalbuf_end('qyz',nchecked(11),nfailed(11),errmax(11),tol)
    call checkvalbuf_end('qzz',nchecked(12),nfailed(12),errmax(12),tol)
#endif
    call update_test_scores(ntests,nfailed,npass)

    deallocate(old_tree)
 endif

 if (id==master) write(*,"(/,a,/)") '<-- KDTREE TEST COMPLETE'

end subroutine test_kdtree
end module testkdtree
